functor JoinTreeFn(structure JTI : JOIN_TREE_INNARDS )
	:> JOIN_TREE where GModel = JTI.GModel 
                     and IP = JTI.IP
		     and RVar = JTI.RVar 
		     and type pot = JTI.pot =
struct

structure JTI = JTI
structure GModel = JTI.GModel
structure IP = JTI.IP
structure RVar = JTI.RVar
structure UGraph = GModel.UGraph 
structure DGraph = GModel.DGraph 
structure RVSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)
structure IntSet = Util.IntSet
structure IntSetSet = Util.IntSetSet

exception Error of string

type pot = JTI.pot

(* A Join tree is represented recursively as a clique potential and a
  list of seperator potenials and thier corrisponding (sub)
  jtrees. This makes the root explicit, although new roots can be had
  by a simple rotation. Leaf nodes will (of course) have a null list
  of decendents.  *)
datatype jtree = Clique of pot * ((pot * jtree) list)
			   
fun size (Clique (_, sep_childs)) =
    let
	val (_,childs) = ListPair.unzip sep_childs
    in
	List.foldl (fn (c,acc)=> size c + acc) 1 childs
    end

fun scope jtree =
    let
	fun f (Clique (cpot, sep_subtrees)) scope =
	    let
		val (_,subtrees) = ListPair.unzip sep_subtrees
		val scope = RVSet.addList(scope, JTI.scope cpot)
		val scope = List.foldl (fn (jtree, scope)=>
					  RVSet.union(scope,
						       (f jtree scope)))
				      scope subtrees
	    in
		scope
	    end
    in
	RVSet.listItems (f jtree RVSet.empty)
    end
	    

fun toDotString jtree =
    let
	val header = "digraph {\n"
	val footer = "}\n"
	fun cliqueToString cpot = "\"" ^ 
				  (List.foldl (fn (v,acc)=> acc ^" " 
							    ^ (RVar.toString v))
					      "" (JTI.scope cpot))
				  ^ "\""
	fun subTree (Clique (cpot, children)) =
	    let
		val (spots,childs) = ListPair.unzip children
		val child_cpots = List.map (fn (Clique (c,_))=> c) childs
		val clique_string = cliqueToString cpot
		val children_edges =
		    List.foldl (fn (c,acc)=> 
				   acc ^ "\n" ^ clique_string 
				   ^ " -> " ^ (cliqueToString c))
			       "" child_cpots
		val rest = List.foldl (fn (c,acc) => acc ^ (subTree c))
				      "" childs
	    in
		children_edges ^ rest
	    end

	val graph = subTree jtree
    in
	header ^ graph ^ footer
    end

(****************************
 Graph Manipulation Functions :  operate on the u/d-graph
 *****************************)

(* return the edges needed to fully connect v to its neighbors in g *)
fun inducedEdges g v =
    let
	val neighbors = UGraph.getNeighbors g v
	fun edgesNeeded v =
	    List.foldr (fn (v',acc)=>if v=v' orelse 
					UGraph.edgeExists g (v,v')
				     then acc 
				     else (v,v')::acc)
		       [] neighbors
		       
	val edges_needed = List.foldl (fn (n,acc) => (edgesNeeded n)@acc)
				      [] neighbors
    in
	edges_needed
    end

(* greedyPick : graph -> vertex
   Given a variable ugraph, pick the vertex to remove which induces the 
   minimal amount edges breaking ties with minimum clique size *)
fun greedyPick g =
    let
	(* evaluate : vertex -> (edges_needed*clique_weight) *)
	fun evaluate v =
	    let
		val neighbors = UGraph.getNeighbors g v
		val edges_needed = inducedEdges g v
		val vars = List.map (UGraph.getVertex g) (v::neighbors)
		val clique_weight = List.foldl (fn (v, acc)=>RVar.dim v * acc)
						1 vars
	    in
		(edges_needed, clique_weight)
	    end
		
	val vs = List.tabulate (UGraph.numVertices g, fn x=>x)
	(* Make sure we don't look at verticies that have been disconnected *)
	val vs = List.filter (fn v=>UGraph.getNeighbors g v <> []) vs 
	val evs = List.map evaluate vs
	val v_evs = ListPair.map (fn (v,(e,s))=>(v,e,s)) (vs,evs) 

	val (pick,_,_) =
	    Util.findMin (fn ((_,e1s,s1),(_,e2s,s2)) =>
			     case Int.compare(List.length e1s, List.length e2s)
			      of GREATER => false
			       | LESS => true
			       | EQUAL => case Int.compare(s1,s2) 
					   of GREATER => false
					    | _ => true)
			 v_evs
    in
	pick
    end

(* elimOrderGreedy : graph -> elim order 

  NOTE: This re-evaluates the weights of all nodes after each pick. A
  more efficient implemetation would only re-evaluate nodes where
  edges are added. *)
fun elimOrderGreedy g =
    if UGraph.getEdges g = [] then []
    else
	let
	    val v = greedyPick g
	    val g' = List.foldl (fn (e,g) => UGraph.addEdge g e)
				(UGraph.disconnectVertex g v)
				(inducedEdges g v)
	in
	    v :: (elimOrderGreedy g')
	end

(* triangulate : variable ugraph -> elim order -> 
		 (new graph, cliques, fill-in edges) *)
fun triangulate ugraph order =
    let
(* 	val _ = print "triangulate\n" *)
	fun tri g [] (clique_set, edges) = (clique_set, edges)
	  | tri g (v::vs) (clique_set, edges) = 
	    let
		val clique = v :: (UGraph.getNeighbors g v)
		val clique = IntSet.addList(IntSet.empty, clique)
		val clique_set' = 
		    if IntSetSet.exists (fn s=>IntSet.isSubset (clique, s))
					clique_set
		    then clique_set
		    else IntSetSet.add (clique_set, clique)
		val new_edges = inducedEdges g v
		val g' = List.foldl (fn (e,g) => UGraph.addEdge g e)
				    (UGraph.disconnectVertex g v)
				    new_edges
		val edges' = new_edges @ edges

	    in
		tri g' vs (clique_set', edges')
	    end

	val (clique_set,edges) = tri ugraph order (IntSetSet.empty, [])
	val clique_set = List.map IntSet.listItems 
				  (IntSetSet.listItems clique_set)
	val clique_set = List.map (fn c=> List.map 
					  (fn v=>UGraph.getVertex ugraph v)
					  c)
			 clique_set
	val new_graph = List.foldr (fn (e,g) => UGraph.addEdge g e)
				   ugraph edges
    in
	(new_graph, clique_set, edges)
    end

(* cliqueTreeFromMaxCliques : maximal cliques -> clique tree *)
fun cliqueTreeFromMaxCliques cliques =
    let
(* 	val _ = print "cliqueTreeFromMaxCliques (max spanning)\n" *)

	(* This could be computed once and stored in an array *)
	fun w (c1, c2) = 
	    List.foldr (fn (v,acc)=> if List.exists (fn x=>v=x) c2
				     then acc+1
				     else acc)
		       0 c1

	val clique_graph = UGraph.fullyConnect (UGraph.make cliques)
    in
	UGraph.maxSpanningTree clique_graph w
    end


(***********************************
 Join Tree Constructions 
************************************)

(* 
  Construct a jtree from a list of potentials and a ugraph (assumed to
  be a tree) of cliques (var lists) 
*)
fun buildJtree pots ctree =
    let
	fun cliqueToString c = List.foldl (fn (c,acc)=> acc ^ " " ^ 
							(RVar.toString c))
				  "" c
(* 	val _ = print "buildJtree\n" *)
	(*val _ = Global.showDot (UGraph.toDotString ctree cliqueToString) *)
		
	(* get the potentials which can be put in this clique 
           and those which must go elsewhere *)
	fun getPots c pots =
	    let
		val scope = RVSet.addList(RVSet.empty,
					   (UGraph.getVertex ctree c))
			    
		fun potScope p = RVSet.addList(RVSet.empty,
						IP.scope p)
	    in
		List.partition (fn p => RVSet.isSubset(potScope p, scope))
			       pots
	    end

	(* recursibly build up a subtree. return the tree and the 
           unused potentials *)
	fun subTree c pots marked =
	    let
		val clique = UGraph.getVertex ctree c
		val clique_set = RVSet.addList(RVSet.empty, clique)
		val nbs = UGraph.getNeighbors ctree c
		val unvisited_nbs = List.filter 
				    (fn x => 
					not (List.exists (fn y=>y=x) marked))
				    nbs
		val (these_pots, rest_pots) = getPots c pots

		val clique_pot = List.foldl JTI.addPotToClique
					    (JTI.initialClique clique)
					    these_pots

		(* fuction to fold over children and build up the subtree *)
		fun f (id, (trees, pots)) = 
		    let 
			val (tree, pots') = subTree id pots (c::marked)
			val child_set = 
			    RVSet.addList(RVSet.empty,
					   (UGraph.getVertex ctree id))
			val sep_scope = RVSet.listItems
					    (RVSet.intersection
					    (clique_set, child_set))
		    in 
			((JTI.initialSep sep_scope, tree)::trees, pots') 
		    end

		(* build the children*)
		val (children, unused_pots) = List.foldl f ([], rest_pots)
							 unvisited_nbs
	    in
		(Clique (clique_pot, children), unused_pots)
	    end

(* 	val _ = print "subTree\n" *)

	val (jtree, leftovers) = subTree 0 pots []
    in
	if not (List.null leftovers) 
	then raise Error "leftover potentials"
	else jtree
    end

(* Putting it together to build the inconsistent jtree *)
fun build gmodel  = 
    let
	val (ipots,moralized) = 
	    case gmodel of GModel.BNet (ipots, dgraph)
			   => (ipots, UGraph.moralize dgraph)
			 | GModel.MNet (ipots, ugraph) => (ipots, ugraph)
	val elim_order = elimOrderGreedy moralized
	val (tg, cliques, edges) = triangulate moralized elim_order
	val clique_tree = cliqueTreeFromMaxCliques cliques
	val jtree = buildJtree ipots clique_tree
    in
	jtree
    end

end
