structure Table :> TABLE =
struct

structure V=Vector

exception Error of string

(* table = dimensions * vector *)
type table = (int list*real V.vector)

fun make dim elist =
    let
	val size = List.foldr op* 1 dim
	val _ = if size <> List.length elist
		then raise Error "make: bad spec"
		else ()
    in
	(dim, V.fromList elist)
    end

fun size (dim,a) = V.length a

fun getValue val_list (dim,a) = 
    let
	val _ = ListPair.foldr (fn (v,d,_)=>if v >= d orelse v < 0 
					    then raise Error "get: bad spec"
					    else ())
			       () (val_list,dim)
		
	fun getIndex [] [] = 0
	  | getIndex (v::vs) (d::ds) = v + d * (getIndex vs ds)

	val index = getIndex val_list dim
    in
	V.sub (a,index)
    end


fun multiply ((d1,a1), (d2,a2)) =
    let
	val _ = if d1 <> d2 then raise Error "mult: mismatched dimensions"
		else ()
	val l1 = V.length a1
	val l2 = V.length a2
	val _ = if l1 <> l2 then raise Error "mult: mismatched vector lengths" 
		else ()

	val a' = V.tabulate (l1, (fn i => Real.*(V.sub(a1,i), 
						 V.sub(a2,i))))
    in
	(d1, a')
    end

fun divide ((d1,a1), (d2,a2)) =
    let
	val _ = if d1 <> d2 then raise Error "div: mismatched dimensions"
		else ()
	val l1 = V.length a1
	val l2 = V.length a2
	val _ = if l1 <> l2 then raise Error "div: mismatched vector lengths" 
		else ()

	val a' = V.tabulate (l1, (fn i => Real./(V.sub(a1,i), 
						 V.sub(a2,i) )))
    in
	(d1, a')
    end


(* reshape : format list -> table -> reshaped table  
   - A one in the format list means use the existing dimension.
     Otherwise, a new dimesion is created, and the existing table values 
     are used.
 *)
fun reshape fnew ([],a) = (* special case -- scalar vector *)
    let
	val _ = if V.length a <> 1 then raise Error "bad scalar vector" 
		else ()
	val _ = List.map (fn x=> if x=1 
				 then raise Error "bad scalar vector" 
				 else ()) fnew
	val len' = List.foldr op* 1 fnew
	val a' = V.tabulate (len', (fn i => V.sub(a, 0)))
    in
	(fnew,a')
    end
  | reshape fnew (dim, a) =
    let
	val (dim',[]) = List.foldr (fn (f,(acc, dls)) => 
				       if f=1
				       then ((hd dls)::acc, (tl dls))
				       else (f::acc, dls))
				  ([], List.rev dim) fnew
	(* convert from new dimesion to old *)
	fun new2old i = 
	    let
		(* expanded index to the *full* base *)
		fun expand _ [] = [] 
		  | expand n (d::ds) = n mod d :: (expand (n div d) ds)
		(* select the dimenions from the old base *)
		fun select [] [] = []
		  | select (ei::eis) (f::fs) = 
		    if f = 1 then ei :: select eis fs
		    else select eis fs
		(* compress an expanded index into the old base *)
		fun compress [ei] [d] = ei (* should be less than d *)
		  | compress (ei::eis) (d::ds) = ei + d * compress eis ds

		val expanded_index = expand i dim'
		val expanded_old_index = select expanded_index fnew
		val old_index = compress expanded_old_index dim
	    in
		old_index
	    end

	val len' = List.foldr op* 1 dim'
	val a' = V.tabulate (len', (fn i => V.sub(a, new2old i)))
    in
	(dim',a')
    end
	handle match => raise Error "reshape: bad specification"

fun sum (_,a) = V.foldr op+ 0.0 a

fun scale (dim,a) c = 
    let 
	val a' = V.tabulate ((V.length a), (fn i => Real.*(V.sub (a,i), c)))
    in
	(dim,a')
    end

(* sumv : format list -> table -> reduced_table 
   - A one in the format list means *keep* this dimension.
   Otherwise, it is summed out.
 *)			      
fun sumv fnew (dim, a) =
    let
	val (dim',[]) = List.foldr (fn (f,(acc, dls)) => 
				       if f=1
				       then ((hd dls)::acc, (tl dls))
				       else (acc,(tl dls)))
				   ([], List.rev dim) fnew

	(* convert from new dimesion to old *)
	fun new2old i = 
	    let
		(* expanded index in the original *)
		fun expand 0 [] = [] 
		  | expand n (d'::ds') = n mod d' :: (expand (n div d') ds')
		(* select the dimenions from the old base *)
		fun select _ [] [] = [[]]
		  | select (ei::eis) (f::fs) (d::ds) = 
		    if f = 1 then List.map (fn x => ei::x) (select eis fs ds)
		    else 
			let
			    val rest = select (ei::eis) fs ds
			in
			    List.foldr 
			    (fn (v,acc)=>(List.map (fn x=> v::x) rest) @ acc)
				       [] (List.tabulate (d, fn x=>x))
			end
		  | select [] (f::fs) (d::ds) = 
		    if f = 1 then raise Error "sumv: no more"
		    else 
			let
			    val rest = select [] fs ds
			in
			    List.foldr 
			    (fn (v,acc)=>(List.map (fn x=> v::x) rest) @ acc)
			               [] (List.tabulate (d, fn x=>x))
			end

		(* compress an expanded index into the old base *)
		fun compress [ei] [d] = ei (* should be less than d *)
		  | compress (ei::eis) (d::ds) = ei + d * compress eis ds
	     val expanded_index = expand i dim'
	     val old_expaned_indices = select expanded_index fnew dim 
	     val old_indices = List.map (fn x=> compress x dim)
					old_expaned_indices
	    in
		old_indices
	    end

	val len' = List.foldr op* 1 dim'
	val a' = V.tabulate (len',
			     (fn i=> 
				 List.foldr (fn (j,acc)=> Real.+(V.sub(a,j), 
								 acc))
					    0.0 (new2old i)))
    in
	(dim',a')
    end
	handle match => raise Error "reshape: bad specification"

fun isApproxEqual eps ((dim,a), (dim',a')) = 
    (dim=dim') andalso ListPair.all (fn (e,e') =>Real.<(Real.abs(e-e'),eps) )
				    (V.foldr op:: [] a,
				     V.foldr op:: [] a')
			   

fun toString (dim,a) = "Table dim: " ^
		       (List.foldl (fn (d,acc) => acc ^ " " ^(Int.toString d))
				   "" dim) ^ "\n    elements:" ^
		       (V.foldl (fn (e,acc) => acc ^ " " ^(Real.toString e))
				    "" a)
		       ^ "\n"
    

end
