(* An inference problem *)
signature PROBLEM =
sig
    structure GModel : D_G_MODEL
    structure RVar : D_R_VAR
    sharing RVar = GModel.RVar

    val model : GModel.gmodel
    val queries : RVar.var list list (* list of queries *)
    val evidence_nodes : RVar.var list
end

(* The result of an inference problem *)
signature ANSWER =
sig
    structure IP : D_IPOTENTIAL 

    val build_time : real
    val inference_times : real list	   
    val answers : IP.ipot list
end

(* Run an engine on an inference problem *)
functor RunEngine(structure InfEng : INFERENCE
		  structure Problem : PROBLEM
		  structure IP : D_IPOTENTIAL
		  sharing InfEng.GModel = Problem.GModel
		  and IP = IP = Problem.GModel.IP = InfEng.IP
		  and InfEng.RVar = Problem.RVar) 
	:> ANSWER where IP = InfEng.IP =
struct
structure IP = IP

val tic = Time.now ()
val engine = InfEng.makeEngine(Problem.model,
			       [],
			       Problem.evidence_nodes)
val toc = Time.now ()
val build_time = Time.toReal (Time.-(toc,tic))

fun doQuery q = 
    let
	val tic = Time.now ()
	val answer = InfEng.query engine q
	val toc = Time.now ()
	val inference_time = Time.toReal (Time.-(toc,tic))
    in
	(inference_time, answer)
    end

val (inference_times,
     answers) = ListPair.unzip (List.map doQuery Problem.queries)
		
end

functor RandomProblem(structure GModel : D_G_MODEL) 
	:> PROBLEM where GModel = GModel
                   and   RVar = GModel.RVar = 
struct
structure GModel = GModel
structure RVar = GModel.RVar
structure IP = GModel.IP
structure LF = GModel.LocalFactor
structure RVSet = ListSetFn(struct 
			     type ord_key = RVar.var
			     val compare = RVar.compare 
			     end)

val rand = Random.rand (6,9)
structure RM = RandomGModelFn(structure GModel = GModel
			    structure Params =
			    struct
			    val num_vars = 20
			    val max_var_dim = 3
			    val extra_factors = 0
			    end)
fun randomRVars var_list n =
    let
	fun f 0 vars = []
	  | f n vars = 
	    let
		val i = Random.randRange (0, List.length vars - 1) rand
	    in
		List.nth (vars,i) :: (f (n-1)
					((List.take (vars,i))@ 
					 List.drop (vars,i+1)))
	    end
    in
	f n var_list
    end

val num_queries = 10
val query_size = 5
val model = RM.model
val scope = List.foldl (fn (p,ps) => RVSet.addList(ps,IP.scope p))
		       RVSet.empty (GModel.getIPotentials RM.model)
val scope = RVSet.listItems scope
val queries = List.tabulate (num_queries, (fn _=> randomRVars scope 
							     query_size))

val evidence_nodes = []
end

(* The classic : cloud, sprinker, rain, wet problem *)
structure P1  :> PROBLEM  where GModel = GM.GModel 
                          and RVar = GM.RVar =
struct
   structure GModel = GM.GModel
   structure RVar = GM.RVar
   structure LF = GM.LocalFactor

   val [c,s,r,w] = List.map RVar.make [(1,2),(2,2),(3,2),(4,2)]
   val fc = LF.CPD (LF.CPD.make ([0.5, 0.5], c, []))
   val fr = LF.CPD (LF.CPD.make ([0.8, 0.2, 0.2, 0.8], r, [c]))
   val fs = LF.CPD (LF.CPD.make ([0.5, 0.9, 0.5, 0.1], s, [c]))
   val fw = LF.CPD (LF.CPD.make ([1.0, 0.1, 0.1, 0.01, 0.0, 0.9, 0.9, 0.99], 
				 w, [s,r]))
   val cpds = [fc,fs,fr,fw]
   val model = GModel.make cpds
   val queries = [[c],[s],[r],[w],[c,s],[c,r],[c,w],[s,r],[s,w],[r,w],
		      [c,s,r],[c,s,w],[c,r,w],[s,r,w],[c,s,r,w]]
   val evidence_nodes = []
end


(* The exported structure *)
structure InferenceTester = 
struct

open GM

structure Tester = TesterFn()

val _ = Tester.setName "Inference"

structure A1_NI = RunEngine(structure InfEng = NaiveInfer
			      structure Problem = P1
			      structure IP = IPotential)
structure A1_VE = RunEngine(structure InfEng = VarElimInfer
			      structure Problem = P1
			      structure IP = IPotential)
structure A1_JT = RunEngine(structure InfEng = JoinTreeInfer
			      structure Problem = P1
			      structure IP = IPotential)
val _ = Tester.addTest
	    "test 1 (sprinkler) Naive == VarElim"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A1_NI.answers, A1_VE.answers))
val _ = Tester.addTest
	    "test 1 (sprinkler) Naive == JoinTree"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A1_NI.answers, A1_JT.answers))

structure P2 = RandomProblem(structure GModel = GM.GModel)
val _ = Util.showDot (GModel.toDotString P2.model) 
structure A2_VE = RunEngine(structure InfEng = VarElimInfer
			    structure Problem = P2
			    structure IP = IPotential)
structure A2_JT = RunEngine(structure InfEng = JoinTreeInfer
			    structure Problem = P2
			    structure IP = IPotential)

val _ = Tester.addTest
	    "test 2 (random) VarElim == JoinTree"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A2_VE.answers, A2_JT.answers))

(* structure A1_LZ = RunEngine(structure InfEng = LazyInfer *)
(* 			    structure Problem = P1 *)
(* 			    structure IP = IPotential) *)
(* structure A2_LZ = RunEngine(structure InfEng = LazyInfer *)
(* 			    structure Problem = P2 *)
(* 			    structure IP = IPotential) *)
(* val _ = Tester.addTest *)
(* 	    "test 1 (sprinkler) LAZY" *)
(* 	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10) *)
(* 				   (A1_NI.answers, A1_LZ.answers)) *)
(* val _ = Tester.addTest *)
(* 	    "test 2 LAZY" *)
(* 	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10) *)
(* 				   (A2_JT.answers, A2_LZ.answers)) *)
(* val _ = print ("LZ build: " ^ (Real.toString A2_LZ.build_time) ^ *)
(* 	     "\n   infer (mean): " *)
(* 	     ^ (Real.toString (mean A2_LZ.inference_times)) *)
(* 	     ^ "\n") *)

fun mean ls = (List.foldl op+ 0.0 ls) / (Real.fromInt (List.length ls))

val PASSED_ALL_TESTS = Tester.runTests ()

val _ = print ("VE build: " ^ (Real.toString A2_VE.build_time) ^
	     "\n   infer (mean): "
	     ^ (Real.toString (mean A2_VE.inference_times))
	     ^ "\n")
val _ = print ("JT build: " ^ (Real.toString A2_JT.build_time) ^
	     "\n   infer (mean): "
	     ^ (Real.toString (mean A2_JT.inference_times))
	     ^ "\n")
end
