(* An inference problem *)
signature PROBLEM =
sig
    structure GModel : GMODEL
    structure Var : VAR
    sharing Var = GModel.Var

    val model : GModel.gmodel
    val queries : Var.var list list (* list of queries *)
    val evidence_nodes : Var.var list
end

(* The result of an inference problem *)
signature ANSWER =
sig
    structure IP : IPOTENTIAL 

    val build_time : real
    val inference_times : real list	   
    val answers : IP.ipot list
end

(* Run an engine on an inference problem *)
functor RunEngine(structure InfEng : INFERENCE
		  structure Problem : PROBLEM
		  structure IP : IPOTENTIAL
		  sharing InfEng.GModel = Problem.GModel
		  and IP = IP = Problem.GModel.IP = InfEng.IP
		  and InfEng.Var = Problem.Var) 
	:> ANSWER where IP = InfEng.IP =
struct
structure IP = IP

val tic = Time.now ()
val engine = InfEng.makeEngine(Problem.model,
			       [],
			       Problem.evidence_nodes)
val toc = Time.now ()
val build_time = Time.toReal (Time.-(toc,tic))

fun doQuery q = 
    let
	val tic = Time.now ()
	val answer = InfEng.query engine q
	val toc = Time.now ()
	val inference_time = Time.toReal (Time.-(toc,tic))
    in
	(inference_time, answer)
    end

val (inference_times,
     answers) = ListPair.unzip (List.map doQuery Problem.queries)
		
end

functor RandomProblem(structure GModel : GMODEL) 
	:> PROBLEM where GModel = GModel
                   and   Var = GModel.Var = 
struct
structure GModel = GModel
structure Var = GModel.Var
structure IP = GModel.IP
structure LF = GModel.LocalFactor

val rand = Random.rand (6,9)

val num_queries = 10
val query_size = 3
val num_vars = 10
val factor_size = 4
val num_factors = 10

val var_list = List.map (fn id => Var.make (id, Random.randRange (2,5) rand))
			(List.tabulate (num_vars, (fn x=>x)))
fun randomVars n =
    let
	fun f 0 vars = []
	  | f n vars = 
	    let
		val i = Random.randRange (0, List.length vars - 1) rand
	    in
		List.nth (vars,i) :: (f (n-1)
					((List.take (vars,i))@ 
					 List.drop (vars,i+1)))
	    end
    in
	f n var_list
    end

fun randomReals 0 = []
  | randomReals n = Real.abs (Random.randReal rand) :: (randomReals (n-1))

fun randomFactor () =
    let
	val vars = randomVars factor_size
	val size = List.foldl op* 1 (List.map Var.dim vars)
    in
	LF.makePotential (randomReals size, vars)
    end

val factors = List.map (fn _=> randomFactor ())
	      (List.tabulate (num_factors, fn x=>x))

val model = GModel.make factors
val queries = List.tabulate (num_queries, (fn _=> randomVars query_size))
val evidence_nodes = []
end

(* The classic : cloud, sprinker, rain, wet problem *)
structure P1  :> PROBLEM  where GModel = GM.GModel 
                          and Var = GM.Var =
struct
   structure GModel = GM.GModel
   structure Var = GM.Var
   structure LF = GM.LocalFactor

   val [c,s,r,w] = List.map Var.make [(1,2),(2,2),(3,2),(4,2)]
   val fc = LF.makeCPD([0.5, 0.5], c, [])
   val fr = LF.makeCPD([0.8, 0.2, 0.2, 0.8], r, [c])
   val fs = LF.makeCPD([0.5, 0.9, 0.5, 0.1], s, [c])
   val fw = LF.makeCPD([1.0, 0.1, 0.1, 0.01, 0.0, 0.9, 0.9, 0.99], w, [s,r])
   val cpds = [fc,fs,fr,fw]
   val model = GModel.make cpds
   val queries = [[c],[s],[r],[w],[c,s],[c,r],[c,w],[s,r],[s,w],[r,w],
		      [c,s,r],[c,s,w],[c,r,w],[s,r,w],[c,s,r,w]]
   val evidence_nodes = []
end


(* The exported structure *)
structure InferenceTester = 
struct

open Global;
open GM;

structure Tester = Tester()

val _ = Tester.setName "Inference"

structure A1_NI = RunEngine(structure InfEng = NaiveInfer
			      structure Problem = P1
			      structure IP = IPotential) 
structure A1_VE = RunEngine(structure InfEng = VarElimInfer
			      structure Problem = P1
			      structure IP = IPotential) 
structure A1_JT = RunEngine(structure InfEng = JoinTreeInfer
			      structure Problem = P1
			      structure IP = IPotential) 
val _ = Tester.addTest 
	    "test 1.1"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A1_NI.answers, A1_VE.answers))
val _ = Tester.addTest 
	    "test 1.2"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A1_NI.answers, A1_JT.answers))


structure P2 = RandomProblem(structure GModel = GM.GModel)
structure A2_VE = RunEngine(structure InfEng = VarElimInfer
			    structure Problem = P2
			    structure IP = IPotential)
structure A2_JT = RunEngine(structure InfEng = JoinTreeInfer
			    structure Problem = P2
			    structure IP = IPotential)
val _ = Tester.addTest
	    "test 2"
	    (fn () => ListPair.all (IPotential.isApproxEqual 10e~10)
				   (A2_VE.answers, A2_JT.answers))

fun mean ls = (List.foldl op+ 0.0 ls) / (Real.fromInt (List.length ls))

val PASSED_ALL_TESTS = Tester.runTests ()

val _ = print ("VE build: " ^ (Real.toString A2_VE.build_time) ^
	     "\n   infer (mean): " 
	     ^ (Real.toString (mean A2_VE.inference_times))
	     ^ "\n")
val _ = print ("JT build: " ^ (Real.toString A2_JT.build_time) ^
	     "\n   infer (mean): " 
	     ^ (Real.toString (mean A2_JT.inference_times))
	     ^ "\n")
end
