structure DGraph :> DGRAPH = 
struct
open Util
exception Error of string

(* list of verticies (the position in the list reflects the vertex id) 
   * adjacency matrix *)
type 'a dgraph = ('a list * bool Array2.array)

fun make vl =
    let
	val dim = List.length vl
	val a = Array2.array (dim,dim,false)
    in
	(vl,a)
    end

fun numVertices (vl,a) = List.length vl

fun addVertex (vl,a) v =
    let
	val (rows,cols) = Array2.dimensions a
	val _ = if rows <> cols then raise Error "addVertex: bad a matrix"
		else ()
	val a' = Array2.array (rows+1,cols+1,false)
	val _ = Array2.modifyi 
		    Array2.ColMajor 
		    (fn (ri,ci,_)=> Array2.sub (a,ri,ci))
		    {base=a',row=0,nrows=SOME(rows-1),
		     col=0,ncols=SOME(cols-1)}
    in
	((vl @ [v], a'), rows) 	(* new graph, vertex id *)
    end

fun getVertices (vl,_) = vl
fun getVertex (vl,_) vi = List.nth (vl, vi)

fun getVertexId (vl,_) v =
    case
	List.find (fn (v',id) => v'= v)
		  (ListPair.zip (vl, List.tabulate (List.length vl, fn x=>x)))
     of SOME (_,id) => SOME id
      | NONE => NONE

fun removeVertex (vl,a) vi =
    let
	val vl' = List.take (vl,vi) @ (List.drop (vl,vi+1))
	val size' = List.length vl'
	val a' = Array2.array (size',size', false)
	val _ = Array2.modifyi 
		    Array2.ColMajor
		    (fn (ri,ci,_)=> 
			let
			    val ri' = if ri < vi then ri else ri+1
			    val ci' = if ci < vi then ci else ci+1
			in
			    Array2.sub (a, ri',ci')
			end)
		    {base=a', col = 0, ncols=NONE, row=0, nrows=NONE}
    in
	(vl',a')
    end

fun updateVertex (vl,a) vi v =
    (List.take (vl, vi) @ [v] @ List.drop(vl, vi+1), a)

fun edgeExists (_,a) (v_out, v_in) = Array2.sub (a, v_out, v_in)

fun addEdge (vl,a) (vout_i, vin_i) = 
    let
	val (rows,cols) = Array2.dimensions a
	val _ = if rows <> cols then raise Error "addEdge: bad a matrix"
		else ()
	val _ = if vout_i >= rows orelse vin_i >= rows 
		then raise Error "addEdge: non existant vertex"
		else ()
	val a' = Array2.array (rows,cols,false)
	val _ = Array2.modifyi 
		    Array2.RowMajor
		    (fn (ri,ci,_)=>Array2.sub (a,ri,ci))
		    {base=a', col = 0, ncols=NONE, row=0, nrows=NONE}
	val _ = Array2.update (a',vout_i, vin_i, true)
    in
	(vl,a')
    end

fun getEdges (vl,a) = Array2.foldi Array2.ColMajor 
		      (fn (r,c,a,acc)=> if a then (r,c)::acc else acc)
		      []
		      {base=a,col=0,ncols=NONE,row=0,nrows=NONE}

fun removeEdge (vl,a) (vout_i, vin_i) = 
    let
	val (rows,cols) = Array2.dimensions a
	val _ = if rows <> cols then raise Error "addEdge: bad a matrix"
		else ()
	val _ = if vout_i >= rows orelse vin_i >= rows 
		then raise Error "removeEdge: non existant vertex"
		else ()
	val a' = Array2.array (rows,cols,false)
	val _ = Array2.modifyi 
		    Array2.RowMajor
		    (fn (ri,ci,_)=>Array2.sub (a,ri,ci))
		    {base=a', col = 0, ncols=NONE, row=0, nrows=NONE}
	val _ = Array2.update (a',vout_i, vin_i, false)
    in
	(vl,a')
    end

fun addParents dgraph (child, parents) =
    List.foldl (fn (parent,g) => addEdge g (parent,child))
	       dgraph
	       parents

fun getChildren (vl,a) vi = 
    let
	val (rows,cols) = Array2.dimensions a

	val _ = if vi >= rows then raise Error "getChildren: bad index"
		else ()

	val row_v = Array2.row (a,vi)
    in
	Vector.foldri (fn (i,a,acc) => if a then i::acc else acc)
		      [] row_v
    end

fun getParents (vl,a) vi = 
    let
	val (rows,cols) = Array2.dimensions a
			  
	val _ = if vi >= cols then raise Error "getParents: bad index"
		else ()
		     
	val col_v = Array2.column (a,vi)
    in
	Vector.foldli (fn (i,a,acc) => if a then i::acc else acc)
		      [] col_v
    end

fun disconnectVertex g v = List.foldl (fn (n,g)=> removeEdge g (v,n)) 
				      g 
				      (getChildren g v)

fun toString (vl,a) v2string =
    let
	fun printVertex id = "(" ^ (Int.toString id) ^ ", " ^
			     (v2string (getVertex (vl,a) id)) ^ "), "
    in
    "DGraph : \n  Vertices: " 
    ^ (String.concat (List.map printVertex 
			       (List.tabulate (List.length vl, fn x=>x))))
    ^ "\n Edges: "
    ^ (String.concat (List.map (fn (vout,vin)=> 
				   "(" ^ (Int.toString vout) ^ "->"
				   ^ (Int.toString vin) ^ ") ")
			       (getEdges (vl,a))))
    ^ "\n"

    end

fun toDotString ugraph nodePrinter =
let
    val header = "digraph {\n"

    val edges = List.foldl (fn ((x,y),acc)=> acc ^ (Int.toString x)
					     ^ " -> "
					     ^ (Int.toString y) ^ ";\n")
			   ""
			   (getEdges ugraph)
    
    val node_labels = List.foldl 
		     (fn (n,acc) => acc ^ (Int.toString n) ^ " [ label = \"" 
				    ^ (nodePrinter (getVertex ugraph n)) 
				    ^ "\" ];\n" ) 
		     ""  (List.tabulate (numVertices ugraph,fn x=>x))
    val footer = "}\n"
in
    header ^ edges ^ node_labels ^ footer
end

fun reachable dgraph (start,goal) =
    let
	fun search current path visited =
	    if current = goal then SOME(path)
	    else 
		let
		    val nbs = IntSet.addList(IntSet.empty,
					     getChildren dgraph current)
		    val unvisited = IntSet.difference(nbs, visited)
		    val visited' = IntSet.add (visited, current)
		in
		    List.foldl (fn (_,SOME(acc)) => SOME(acc)
				 | (c,NONE) => search c (c::path) visited')
			       NONE
			       (IntSet.listItems unvisited)
		end
    in
	search start [] IntSet.empty
    end
end
