
structure UGraph :> UGRAPH =
struct

structure DGraph = DGraph

type 'a ugraph = 'a DGraph.dgraph  (* carefully use the directed graph rep *)

val make = DGraph.make

fun fromDGraph dgraph =
    let
	val edges = DGraph.getEdges dgraph
	val new_edges = List.map (fn (a,b)=>(b,a)) edges
    in
	List.foldr (fn (e,g)=>DGraph.addEdge g e) dgraph new_edges
    end

val numVertices = DGraph.numVertices
val addVertex = DGraph.addVertex
val getVertex = DGraph.getVertex
val getVertexId = DGraph.getVertexId
val getVertices = DGraph.getVertices
val removeVertex = DGraph.removeVertex
val updateVertex = DGraph.updateVertex
val toString = DGraph.toString

fun edgeExists g (v1,v2) = DGraph.edgeExists g (v1, v2) orelse
			   DGraph.edgeExists g (v2, v1) 

fun addEdge g (v1,v2) = DGraph.addEdge (DGraph.addEdge g (v1,v2)) (v2,v1)

fun connect g [] = g
  | connect g (v::vs) = connect (List.foldl (fn (v',g) => addEdge g (v,v')) 
					    g vs)
				vs

val getEdges =  DGraph.getEdges (* This returns two pair for each edge *)

fun removeEdge g (v1,v2) = DGraph.removeEdge (DGraph.removeEdge g (v1,v2)) (v2,v1)

fun getNeighbors g v = DGraph.getChildren g v

fun disconnectVertex g v = List.foldl (fn (n,g)=> removeEdge g (v,n)) 
				      g 
				      (getNeighbors g v)
fun moralize dgraph =
    let
	val v_ids = List.tabulate (DGraph.numVertices dgraph, (fn x=>x))
	val plist = List.map (fn v=> DGraph.getParents dgraph v) v_ids
	val ug = fromDGraph dgraph

	fun newEdges [] = []
	  | newEdges (p::ps) = (p,ps) :: newEdges ps
			       
	val new_edges = List.map newEdges plist

	val massMarry = List.foldl 
			(fn ((p,ps),g) => List.foldl 
					      (fn (p',g)=> addEdge g (p,p'))
					      g ps)
    in
	List.foldl (fn (coparents,g)=>massMarry g coparents)
		   ug new_edges
    end

fun maxSpanningTree ugraph w =
    let
	val connected_sets = List.map (fn v=>[v]) 
				      (List.tabulate (numVertices ugraph, fn x=>x))
	val sorted_edges = ListMergeSort.sort 
			   (fn ((v1,v2),
				(v1',v2')) => w(getVertex ugraph v1, 
						getVertex ugraph v2) 
					      < 
					      w(getVertex ugraph v1',
						getVertex ugraph v2'))
					      (getEdges ugraph)

	fun findV e (cs::css) = if List.exists (fn x=>x=e) cs then (cs,css)
				else let val (s,rest) = findV e css 
				     in (s, cs::rest) end (*non-tail*)

	fun pick _ [] = []
	  | pick c_sets ((a,b)::es) =
	    let
		(* find the sets containing a and b *)
		val (a_set,a_rest) = findV a c_sets
		val (b_set,b_rest) = findV b c_sets
	    in
		if a_set = b_set then pick c_sets es
		else 
		    let 
			val new_c_sets = (a_set @ b_set) ::
					 (List.filter (fn s=>not (s=b_set)) a_rest)
		    in
			(a,b) :: (pick new_c_sets es)
		    end
	    end

	val edges = pick connected_sets sorted_edges
	val g' = make (getVertices ugraph)
	val g' = List.foldr (fn (e,g)=> addEdge g e) g' edges
    in
	g'
    end

fun test () = (* from CLRS pg 668 *)
    let
	val g = make ["a", "b", "c", "d", "e", "f", "g", "h", "i"]
	val g = List.foldr (fn (e,g)=>addEdge g e) g
		[ (0,1), (0, 7), (1,2), (1,7), (2,3), (2,5), (2,8), (3,4), 
		  (3,5), (4,5), (5,6), (6,7), (6,8), (7,8)]
	fun w ("a","b") = 4
	  | w ("a","h") = 8
	  | w ("b","c") = 8
	  | w ("b","h") = 11
	  | w ("c","d") = 7
	  | w ("c","f") = 4
	  | w ("c","i") = 2
	  | w ("d","e") = 9
	  | w ("d","f") = 14
	  | w ("e","f") = 10
	  | w ("f","g") = 2
	  | w ("g","h") = 1
	  | w ("g","i") = 6
	  | w ("h","i") = 7
	  | w (a,b) = w(b,a)
    in
	maxSpanningTree g w
    end

fun fullyConnect ugraph = 
    let
	fun f g [] = g
	  | f g (v::vs) = f (List.foldr (fn (v',g)=> addEdge g (v,v')) g vs) vs
    in
	f ugraph (List.tabulate (numVertices ugraph, fn x=>x))
    end

fun toDotString ugraph nodePrinter =
let
    val header = "graph {\n"

    val edges = List.foldl (fn ((x,y),acc)=> acc ^ (Int.toString x)
					     ^ " -- "
					     ^ (Int.toString y) ^ ";\n")
			   ""
			   (List.filter (fn (x,y)=> x<y) (getEdges ugraph))
    
    val node_labels = List.foldl 
		     (fn (n,acc) => acc ^ (Int.toString n) ^ " [ label = \"" 
				    ^ (nodePrinter (getVertex ugraph n)) 
				    ^ "\" ]\n")
		     ""  (List.tabulate (numVertices ugraph,fn x=>x))
    val footer = "}\n"
in
    header ^ edges ^ node_labels ^ footer
end


end
